/*
  PokeMini - Pokmon-Mini Emulator
  Copyright (C) 2009-2012  JustBurn

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <kos.h>

#include "PokeMini.h"
#include "Hardware.h"
#include "Video_x3.h"
#include "PokeMini_BG3.h"
#include "Joystick.h"

const char *AppName = "PokeMini " PokeMini_Version " Dreamcast";

KOS_INIT_FLAGS(INIT_DEFAULT);

// --------

int emurunning = 1;
unsigned short *VIDEO;
int PixPitch, ScOffP;
snd_stream_hnd_t SoundStreamH;

#define DC_CONT_TRG_DEAD	80
#define DC_CONT_JOY_DEAD	40
#define CONT_LTRIG	0x10000
#define CONT_RTRIG	0x20000

int VideoMode(int pal);

// Joystick names and mapping (NEW IN 0.5.0)
char *DCJoy_KeysNames[] = {
	"Off",		// -1
	"C",		// 0
	"B",		// 1
	"A",		// 2
	"Start",	// 3
	"DPad Up",	// 4
	"DPad Down",	// 5
	"DPad Left",	// 6
	"DPad Right",	// 7
	"Z",		// 8
	"Y",		// 9
	"X",		// 10
	"D",		// 11
	"DPad2 Up",	// 12
	"DPad2 Down",	// 13
	"DPad2 Left",	// 14
	"DPad2 Right",	// 15
	"L Trigger",	// 16
	"R Trigger"	// 17
};
int DCJoy_KeysMapping[] = {
	10,		// Menu
	1,		// A
	2,		// B
	17,		// C
	4,		// Up
	5,		// Down
	6,		// Left
	7,		// Right
	3,		// Power
	16		// Shake
};

// Custom command line (NEW IN 0.5.0)
int clc_tvout_pal = 0;
const TCommandLineCustom CustomConf[] = {
	{ "tvout_pal", &clc_tvout_pal, COMMANDLINE_BOOL },
	{ "", NULL, COMMANDLINE_EOL }
};

// Platform menu (REQUIRED >= 0.4.4)
int UIItems_PlatformC(int index, int reason);
TUIMenu_Item UIItems_Platform[] = {
	PLATFORMDEF_GOBACK,
	{ 0,  1, "TV Out: %s", UIItems_PlatformC },
	{ 0,  9, "Define Joystick...", UIItems_PlatformC },
	PLATFORMDEF_SAVEOPTIONS,
	PLATFORMDEF_END(UIItems_PlatformC)
};
int UIItems_PlatformC(int index, int reason)
{
	int videochanged = 0;
	if (reason == UIMENU_OK) reason = UIMENU_RIGHT;
	if (reason == UIMENU_CANCEL) UIMenu_PrevMenu();
	if (reason == UIMENU_LEFT) {
		switch (index) {
			case 1: // TV Out
				clc_tvout_pal = !clc_tvout_pal;
				videochanged = 1;
				break;
		}
	}
	if (reason == UIMENU_RIGHT) {
		switch (index) {
			case 1: // TV Out
				clc_tvout_pal = !clc_tvout_pal;
				videochanged = 1;
				break;
			case 9: // Define keys
				JoystickEnterMenu();
				break;
		}
	}
	if (videochanged) VideoMode(clc_tvout_pal);
	UIMenu_ChangeItem(UIItems_Platform, 1, "TV Out: %s", clc_tvout_pal ? "PAL" : "NTSC");
	return 1;
}

// Set video
int VideoMode(int pal)
{
	if (vid_check_cable() == CT_VGA) {
		// VGA
		vid_set_mode(DM_320x240_VGA, PM_RGB565);
	} else {
		// Composite or RGB
		if (pal) {
			vid_set_mode(DM_320x240_PAL, PM_RGB565);
		} else {
			vid_set_mode(DM_320x240_NTSC, PM_RGB565);
		}
	}
	PixPitch = 320;
	vid_border_color(0, 0, 0);
	memset(vram_s, 0, 1024*1024*2);
	VIDEO = vram_s + (24 * PixPitch) + 16;

	return 1;
}

// Handle keys
void HandleKeys()
{
	maple_device_t *cont;
	cont_state_t *state;
	uint32_t keys;

	cont = maple_enum_type(0, MAPLE_FUNC_CONTROLLER);
	if (!cont) return;
	state = (cont_state_t *)maple_dev_status(cont);
	if (!state) return;

	JoystickAxisEvent(0, state->joyx);
	JoystickAxisEvent(1, state->joyy);

	keys = (state->buttons & 0xFFFF)
	 | (state->ltrig > DC_CONT_TRG_DEAD ? CONT_LTRIG : 0)
	 | (state->rtrig > DC_CONT_TRG_DEAD ? CONT_RTRIG : 0);
	JoystickBitsEvent(keys);
}

// Emulator sound
int16_t soundsamp[SND_STREAM_BUFFER_MAX/8 + 1024];
void *emulatorsound(snd_stream_hnd_t hnd, int smp_req, int *smp_recv)	// smp_* is actually bytes?
{
	MinxAudio_GenerateEmulatedS16(soundsamp, smp_req>>1, 1);
	*smp_recv = smp_req;
	return soundsamp;
}

enum {
	soundstate_idle,
	soundstate_start,
	soundstate_playing,
	soundstate_stop,
	soundstate_end,
	soundstate_endok,
};
volatile int emulatorsoundstate = soundstate_idle;
static semaphore_t *emulatorsound_sem = NULL;
void emulatorsoundthr(void *ptr)
{
	snd_stream_hnd_t SoundStreamH;
	emulatorsound_sem = sem_create(0);

	snd_stream_init();
	SoundStreamH = snd_stream_alloc(NULL, SND_STREAM_BUFFER_MAX/8);
	while (emulatorsoundstate != soundstate_end) {
		switch(emulatorsoundstate) {
			case soundstate_idle:
				sem_wait(emulatorsound_sem);
				break;
			case soundstate_start:
				snd_stream_reinit(SoundStreamH, emulatorsound);
				snd_stream_start(SoundStreamH, MINX_AUDIOFREQ, 0);
				snd_stream_volume(SoundStreamH, 255);
				emulatorsoundstate = soundstate_playing;
				break;
			case soundstate_playing:
				if (snd_stream_poll(SoundStreamH) < 0) {
					snd_stream_stop(SoundStreamH);
					snd_stream_reinit(SoundStreamH, emulatorsound);
					snd_stream_start(SoundStreamH, MINX_AUDIOFREQ, 0);
					snd_stream_volume(SoundStreamH, 255);
				}
				break;
			case soundstate_stop:
				if (emulatorsoundstate != soundstate_idle) snd_stream_stop(SoundStreamH);
				emulatorsoundstate = soundstate_idle;
				break;
		}
		thd_pass();
	}
	snd_stream_destroy(SoundStreamH);
	snd_stream_shutdown();
	emulatorsoundstate = soundstate_endok;
}

// Enable / Disable sound
void enablesound(int sound)
{
	static int lastsound = 0;
	MinxAudio_ChangeEngine(sound);
	if (sound & !lastsound) {
		// Enabling
		emulatorsoundstate = soundstate_start;
		sem_signal(emulatorsound_sem);
		while(emulatorsoundstate != soundstate_playing) thd_pass();
	}
	if (!sound & lastsound) {
		// Disabling
		emulatorsoundstate = soundstate_stop;
		sem_signal(emulatorsound_sem);
		while(emulatorsoundstate != soundstate_idle) thd_pass();
	}
	lastsound = sound;
}

// Menu loop
void menuloop()
{
	uint16_t *VIDEOOFF;

	// Stop sound
	enablesound(0);

	// Create offscreen buffer
	VIDEOOFF = (uint16_t *)malloc(PixPitch*192*2);
	memset(VIDEOOFF, 0, PixPitch*192*2);

	// Update EEPROM
	UIMenu_SaveEEPDisplay_16((uint16_t *)VIDEOOFF, PixPitch);
	memcpy(VIDEO, VIDEOOFF, PixPitch*192*2);
	PokeMini_SaveFromCommandLines(0);

	// Menu's loop
	while (emurunning && (UI_Status == UI_STATUS_MENU)) {
		// Delay some time
		thd_sleep(16);

		// Handle keys
		HandleKeys();

		// Screen rendering
		UIMenu_Display_16((uint16_t *)VIDEOOFF, PixPitch);

		// Copy offscreen to display
		memcpy(VIDEO, VIDEOOFF, PixPitch*192*2);
	}

	// Free offscreen buffer and apply configs
	free(VIDEOOFF);
	vid_border_color(0, 0, 0);

	// Apply configs
	PokeMini_ApplyChanges();
	if (UI_Status == UI_STATUS_EXIT) emurunning = 0;
	else if (CommandLine.sound) enablesound(CommandLine.sound);
	PokeMini_VideoBlit(VIDEO, PixPitch);
}

// Main function
int main(int argc, char **argv)
{
	int rumbling_old = 0;
	uint64 time = 0, nexttime = 0;

	// Process no arguments
	printf("%s\n\n", AppName);
	PokeMini_InitDirs(argv[0], NULL);
	CommandLineInit();
	CommandLineConfFile("pokemini.cfg", "pokemini_dc.cfg", CustomConf);
	strcpy(CommandLine.bios_file, "/cd/bios.min");
	strcpy(CommandLine.eeprom_file, "/vmu/PokeMini.eep");
	CommandLine.sound = CommandLine.sound ? 1 : 0;
	JoystickSetup("Dreamcast", 0, DC_CONT_JOY_DEAD, DCJoy_KeysNames, 18, DCJoy_KeysMapping);

	// Allocate sound stream
	emulatorsoundstate = soundstate_idle;
	if (!thd_create(emulatorsoundthr, NULL)) {
		printf("Error creating sound thread\n");
	}

	// Set video spec and check if is supported
	if (!PokeMini_SetVideo((TPokeMini_VideoSpec *)&PokeMini_Video3x3, 16, CommandLine.lcdfilter, CommandLine.lcdmode)) {
		fprintf(stderr, "Couldn't set video spec\n");
		exit(1);
	}
	UIMenu_SetDisplay(288, 192, PokeMini_BGR16, (uint8_t *)PokeMini_BG3, (uint16_t *)PokeMini_BG3_PalBGR16, (uint32_t *)PokeMini_BG3_PalBGR32);

	// Initialize video
	VideoMode(0);

	// Initialize the emulator
	printf("Starting emulator...\n");
	if (!PokeMini_Create(POKEMINI_GENSOUND, 0)) {
		fprintf(stderr, "Error while initializing emulator.\n");
	}

	// Setup palette and LCD mode
	PokeMini_VideoPalette_Init(PokeMini_BGR16, 1);
	PokeMini_VideoPalette_Index(CommandLine.palette, CommandLine.custompal);
	PokeMini_ApplyChanges();

	// Load stuff
	PokeMini_UseDefaultCallbacks();
	if (!PokeMini_LoadFromCommandLines("Using FreeBIOS", "EEPROM data will be discarded!")) {
		UI_Status = UI_STATUS_MENU;
	}

	// Enable sound & init UI
	printf("Running emulator...\n");
	UIMenu_Init();
	enablesound(CommandLine.sound);

	// Emulator's loop
	while (emurunning) {
		// Emulate 1 frame
		PokeMini_EmulateFrame();
		do {
			time = timer_us_gettime64();
		} while (time < nexttime);
		nexttime = time + 13888;

		// Screen rendering
		if (PokeMini_Rumbling || rumbling_old) {
			PokeMini_VideoRect_16(VIDEO, PixPitch, 0, -2, PixPitch, 2, 0x0000);
			PokeMini_VideoRect_16(VIDEO, PixPitch, 0, 192, PixPitch, 2, 0x0000);
			rumbling_old = PokeMini_Rumbling;
		}
		if (LCDDirty) {
			if (PokeMini_Rumbling) {
				PokeMini_VideoBlit(VIDEO + PokeMini_GenRumbleOffset(PixPitch), PixPitch);
			} else {
				PokeMini_VideoBlit(VIDEO, PixPitch);
			}
			LCDDirty--;
		}

		// Handle keys
		HandleKeys();

		// Menu
		if (UI_Status == UI_STATUS_MENU) menuloop();
		thd_pass();
	}

	// Stop sound & free UI
	enablesound(0);
	UIMenu_Destroy();

	// Save Stuff
	PokeMini_SaveFromCommandLines(1);

	// Terminate...
	printf("Shutdown emulator...\n");
	PokeMini_VideoPalette_Free();
	PokeMini_Destroy();

	// Destroy sound stream
	emulatorsoundstate = soundstate_end;
	sem_signal(emulatorsound_sem);
	while(emulatorsoundstate != soundstate_endok) thd_pass();

	// Going back to dctool
	printf("Exit!\n");

	return 0;
}
